; KEYBOARD.ASM               
;
; The assembly language file containing the low level keyboard handler (ISR)
; and the appropriate functions that makes use of the ISR.
;
; This file is written specifically for QBasic programs and can be accessed
; by using QBasic's CALL Absolute() procedure (see QBasic examples).
;
; by Gary Neal, Jr. -- http://garyneal.tripod.com/
;
; Written for and assembled with NASM: The Netwide-Assembly Project.
               
               bits    16              ; This is a DOS 16 bit program

               org     0               ; Machine code and data are paragraph
                                       ; aligned.

; Offsets to the Keyboard interrupt handler functions.

               dw      InstallKeyISR   ; Offset address to InstallKeyISR
               dw      RemoveKeyISR    ; Offset address to RemoveKeyISR
               dw      KeyPressed      ; Offset address to KeyPressed
               dw      KeyPressedNow   ; Offset address to KeyPressedNow
               dw      KeyHardAborted  ; Offset address to KeyHardAborted

oldIsr9        dd      0               ; Storage for BIOS interrupt address
extended       db      0               ; Extended key flag
hardAbort      db      0               ; Hard Abort Flag
chainIsr       db      0               ; Chaining flag
keyIsPressed   resb    128             ; Key is pressed array
keyWasPressed  resb    128             ; Key was pressed array

; KeyboardInterruptHandler
;
; The interrupt handler used to capture the keyboard states.

KeyISR         push    ds              ; Save state of DS
               push    ax              ; Save state of AX
               push    bx              ; Save contents of BX
               sti                     ; Re-enable interrupts immediately
               mov     ax,cs           ; Get the code segment
               mov     ds,ax           ; Set data segment to code segment

; Get the keyboard scan code into AX.

               in      al,60h          ; Get Keyboard scan code into AL.
               xor     ah,ah           ; Zero Most Significant Byte.

; Set the keyboard table to the appropriate state.

               mov     bx,ax           ; Set BX to keyboard scan code
               and     bx,7Fh          ; Keep scan code range within 0-127

; Check for extended key.
               cmp     ax,0E0h         ; Was an extended key pressed?
               jne     TestKey         ; If no, continue with ISR.
               mov     byte [extended],1 ; If yes, flag extended code.
               jmp     CheckChaining   ; Exit Keyboard ISR.


; Test the key being pressed and write appropriate flags.

TestKey        mov     ah,[extended]   ; Get extended key flag into AH.
               mov     byte [extended],0 ; Clear the extended code flag.
               or      al,al           ; Check for Key Released flag.
               jl      KeyReleased     ; If set, branch.

               or      ah,ah           ; Extended key pressed?
               jne     ExtendedPress   ; If yes, branch.

               mov     al,byte [keyIsPressed + bx]
               or      al,1            ; Set normal key pressed flag.
               mov     byte [keyIsPressed + bx],al
               mov     al,byte [keyWasPressed + bx]
               or      al,1            ; Set normal key pressed flag.
               mov     byte [keyWasPressed + bx],al
               jmp     TestReboot      ; Branch.

ExtendedPress  mov     al,byte [keyIsPressed + bx]
               or      al,80h          ; Set extended key pressed flag.
               mov     byte [keyIsPressed + bx],al
               mov     al,byte [keyWasPressed + bx]
               or      al,80h          ; Set extended key pressed flag.
               mov     byte [keyWasPressed + bx],al
               jmp     TestReboot      ; Branch

KeyReleased    or      ah,ah           ; Extended key released?
               jne     ExtendedRel     ; If yes, branch.

               mov     al,byte [keyIsPressed + bx]
               and     al,0FEh         ; Reset normal key flag.
               mov     byte [keyIsPressed + bx],al
               jmp     CheckChaining   ; Exit ISR

ExtendedRel    mov     al,byte [keyIsPressed + bx]
               and     al,7Fh          ; Reset extended key flag.
               mov     byte [keyIsPressed + bx],al
               jmp     CheckChaining   ; Exit ISR

; Test for the Reboot key press sequence.
; If chaining enabled, we must also test for CTRL-C and CTRL-BREAK sequences.

TestReboot     mov     al,byte [keyIsPressed + 1Dh]  ; CTRL Key
               or      al,al                         ; CTRL Key Pressed?
               je      CheckChaining                 ; If not, exit.
               mov     al,byte [chainIsr]            ; Get chain flag.
               or      al,al                         ; Ae we chaining?
               je      RebootOnly                    ; If no, just chk reboot
               mov     bl,byte [keyIsPressed + 2Eh]  ; CTRL-C
               mov     bh,byte [keyIsPressed + 46h]  ; CTRL-BREAK
               and     bh,0FEh                       ; Not Scroll Lock
               or      bh,bl                         ; CTRL-C or CTRL-BREAK?
               jne     RestoreBIOS                   ; Yes, restore BIOS

RebootOnly     mov     ah,byte [keyIsPressed + 38h]  ; ALT Key
               mov     bl,byte [keyIsPressed + 53h]  ; DEL Key
               mov     bh,byte [keyIsPressed + 4Fh]  ; END Key
               or      ah,ah                         ; ALT Key Pressed?
               je      CheckChaining                 ; If not, exit.
               or      bh,bl                         ; DEL or END key pressed?
               je      CheckChaining                 ; If not, exit.

; Restore BIOS Keyboard ISR manually (removing our keyboard handler).

RestoreBIOS    push    es              ; Save state of ES
               xor     ax,ax           ; Zero out AX
               mov     es,ax           ; Set ES to AX
               cli                     ; Disable interrupts.
               mov     bx,word [oldIsr9 + 2]
               mov     ax,word [oldIsr9]
               mov     word [es:38],bx ; Write segment address of BIOS ISR
               mov     word [es:36],ax ; Write offset address of BIOS ISR
               sti                     ; Re-enable interrupts
               pop     es              ; Restore state of ES

; Flag the forced keyboard handler removal.

               xor     ax,ax           ; Zero out AX
               mov     word [oldIsr9 + 2],ax
               mov     word [oldIsr9],ax
               mov     byte [hardAbort],-1
               jmp     ExitKeyISR

; Check to see if we're chaining the interrupts.

CheckChaining  mov     al,byte [chainIsr]
               or      al,al           ; Are we chaining old ISR?
               je      ExitKeyISR      ; If no, exit.
               pop     bx              ; Restore BX register
               pop     ax              ; Restore AX register
               pop     ds              ; Restore data segment
               jmp     far [cs:oldIsr9] ; Transfer control to BIOS's handler

; Clear keyboard controller.

ExitKeyISR     in      al,61h          ; Read keyboard control register
               mov     ah,al           ; Preserve control register state
               or      al,80h          ; Toggle acknowledge bit high
               out     61h,al          ; Write control register
               mov     al,ah           ; Get previous control register state
               out     61h,al          ; Toggle acknowledge bit low

; Acknowledge the interrupt to the PIC.

               mov     al,20h          ; Get Non-specific EOI code
               out     20h,al          ; Write code to PIC

; Clean up and exit.

               pop     bx              ; Restore value of BX.
               pop     ax              ; Restore value of AX.
               pop     ds              ; Restore value of DS.
               iret                    ; Return from Interrupt.

; InstallKeyISR
;
; Sets the keyboard interrupt vector to point to our ISR.

InstallKeyISR  push    bp              ; Save base pointer
               mov     bp,sp           ; Create a new pointer
               push    ds              ; Save state of DS
               push    bx              ; Save state of BX
               push    ax              ; Save state of AX

; Reset the keyboard handler flags.

               mov     bx,[bp + 6]         ; Get address to flag parameter
               mov     ax,[bx]             ; Get chain flag parameter
               mov     bx,cs               ; Get current segment from CS
               mov     ds,bx               ; Set DS to current segment
               mov     byte [hardAbort],0  ; No hard abort
               mov     byte [extended],0   ; No extended key pressed
               mov     byte [chainIsr],al  ; Set flag to chaining parameter

; Clear the contents of the keyboard arrays.

               push    cx              ; Save CX on stack
               push    es              ; Save ES on stack
               push    di              ; Save direction index on stack
               mov     ax,cs           ; Get code segment
               mov     es,ax           ; Set extra segment to code segment
               mov     di,keyIsPressed ; Set DI to point to arrays
               cld                     ; Set direction flag forward
               xor     ax,ax           ; Set write value to zero
               mov     cx,128          ; Set count to 128 words
               rep     stosw           ; Write 128 words
               pop     di              ; Restore direction index
               pop     es              ; Restore extra segment
               pop     cx              ; Restore count register

; Test to see if the Keyboard Handler is already installed.

               mov     ax,word [oldIsr9 + 2]
               or      ax,word [oldIsr9]
               jne     ExitInstall     ; If installed, exit.

; Get BIOS's keyboard handler address.

               mov     ax,3509h        ; Set AX to function
               push    es              ; Save state of ES
               int     21h             ; Call DOS function
               mov     word [oldIsr9 + 2],es
               mov     word [oldIsr9],bx
               pop     es              ; Restore state of ES

; Set our keyboard handler's address in interrupt table.

               push    dx              ; Save state of DX
               mov     ax,2509h        ; Set AX to DOS function number
               mov     dx,KeyISR       ; Set DX to offset address of ISR
               int     21h             ; Call DOS function
               pop     dx              ; Restore state of DX

ExitInstall    pop     ax              ; Restore state of AX
               pop     bx              ; Restore state of BX
               pop     ds              ; Restore state of DS
               pop     bp              ; Restore base pointer
               retf    2               ; Return to calling program

; RemoveKeyISR
;
; Restores BIOS's keyboard handler.

RemoveKeyISR   push    ds              ; Save state of DS
               push    ax              ; Save state of AX
               mov     ax,cs           ; Get current code segment
               mov     ds,ax           ; Set data segment to code segment

; Check to see if valid vector address was stored.

               mov     ax,word [oldIsr9 + 2]
               or      ax,word [oldIsr9]
               je      ExitRemove      ; If zero store, exit.

; Restore old Interrupt address to vector 9.

               push    dx              ; Save state of DX
               lds     dx,[oldIsr9]    ; Get vector address into DS:DX
               mov     ax,2509h        ; Set AX to DOS function
               int     21h             ; Call DOS function
               pop     dx              ; Restore state of DX

; Clean up and exit.

ExitRemove     pop     ax              ; Restore state of AX
               pop     ds              ; Restore state of DS
               retf                    ; Return to calling function.

; KeyPressed
;
; Checks to see if a key is or had been pressed since previous call.

KeyPressed     push    bp              ; Save the base pointer
               mov     bp,sp           ; Set a new pointer
               push    ax              ; Save register AX
               push    bx              ; Save register BX

; Get the scancode value from BASIC.

               mov     bx,[bp + 6]     ; Get the variable address
               mov     ax,[bx]         ; Get scancode value
               and     ax,7Fh          ; Ensure scancode is valid

; Get the key state from the keyboard arrays.

               mov     bx,ax           ; Set the index
               push    ds              ; Save BASIC's Data Segment
               mov     ax,cs           ; Get the Code Segment
               mov     ds,ax           ; Set Data Segment to Code Segment
               mov     al,byte [keyIsPressed + bx]
               or      al,byte [keyWasPressed + bx]
               mov     byte [keyWasPressed + bx],0
               pop     ds              ; Restore BASIC's Data Segment
               xor     ah,ah           ; Get return value ready

; Set the BASIC variable to the key state.

               mov     bx,[bp + 8]     ; Get variable address
               mov     [bx],ax         ; Set variable value

; Clean up and return.

               pop     bx              ; Restore register BX
               pop     ax              ; Restore register AX
               pop     bp              ; Restore Base Pointer
               retf    4               ; Return and restore stack

; KeyPressedNow
;
; Gets the current keyboard state.

KeyPressedNow  push    bp
               mov     bp,sp
               push    ax
               push    bx

; Get the scancode value from BASIC.

               mov     bx,[bp + 6]     ; Get the variable address
               mov     ax,[bx]         ; Get scancode value
               and     ax,7Fh          ; Ensure scancode is valid

; Get the key state from the keyboard arrays.

               mov     bx,ax           ; Set the index
               push    ds              ; Save BASIC's Data Segment
               mov     ax,cs           ; Get the Code Segment
               mov     ds,ax           ; Set Data Segment to Code Segment
               mov     al,byte [keyIsPressed + bx]
               mov     byte [keyWasPressed + bx],0
               pop     ds              ; Restore BASIC's Data Segment
               xor     ah,ah           ; Get return value ready

; Set the BASIC variable to the key state.

               mov     bx,[bp + 8]     ; Get variable address
               mov     [bx],ax         ; Set variable value

; Clean up and return.

               pop     bx              ; Restore register BX
               pop     ax              ; Restore register AX
               pop     bp              ; Restore Base Pointer
               retf    4               ; Return and restore stack

; KeyHardAborted
;
; Returns the state of the hardAbort variable.

KeyHardAborted push    bp              ; Save the Base Pointer
               mov     bp,sp           ; Set up a new pointer
               push    ax              ; Save state of AX
               push    bx              ; Save state of BX

; Get hard abort state into AX

               push    ds              ; Save BASIC's data segment
               mov     ax,cs           ; Get code segment into AX
               mov     ds,ax           ; Set data segment to AX
               mov     al,byte [hardAbort]
               cbw                     ; Set hard abort state in AX
               pop     ds              ; Restore BASIC's data segment

; Set BASIC's variable to hard abort state

               mov     bx,[bp + 6]     ; Get address of return variable
               mov     [bx],ax         ; Set return value in variable

; Clean up and Return

               pop     bx              ; Restore register BX
               pop     ax              ; Restore register AX
               pop     bp              ; Restore Base Pointer
               retf    2               ; Return and restore stack
